import json
import mimetypes

from typing import Dict, Any
from tpcwj.utils import Bytes


class PayloadMetadata:
    """
    Metadata of payload in host image.

    Attributes
    ----------
    start_at : int
        Where the payload field begin in host image data.

    end_at : int
        Where the payload field end in host image data.

    is_first_slice : bool
        Whether the payload is the first slice.

    full_length : int
        The length of the origin file.

    slice_count : int
        The amount of the slices.

    slice_num : int
        The number of the slice.

    name : str
        The name of the origin file.
        It is empty when is_first_slice set to false.
    """
    start_at: int
    end_at: int
    full_length: int
    is_first_slice: bool
    slice_count: int
    slice_num: int
    name: str

    def __init__(self,
                 start_at: int,
                 end_at: int,
                 full_length: int,
                 is_first_slice: bool,
                 slice_count: int,
                 slice_num: int,
                 name: str) -> None:
        self.start_at = start_at
        self.end_at = end_at
        self.slice_count = slice_count
        self.slice_num = slice_num
        self.is_first_slice = is_first_slice
        self.name = name
        self.full_length = full_length

    @classmethod
    def from_json(cls, json_str: str, metadata_lenth: int):
        metadata = json.loads(json_str)
        start_at = metadata_lenth+1
        end_at = start_at+metadata['length']
        full_length = metadata['full_length']
        is_first_slice = False
        slice_count = metadata['slice']
        slice_num = metadata['count']
        name = ''
        if 'data' in metadata:
            is_first_slice = True
            name = metadata['data'][0]['name']
        return cls(start_at, end_at, full_length, is_first_slice, slice_count, slice_num, name)

    def to_json(self, payload_id) -> str:
        json_metadata: Dict[str, Any]
        json_metadata = dict({
            'id': payload_id,
            'slice': self.slice_count,
            'full_length': self.full_length,
            'count': self.slice_num,
            'length': self.end_at-self.start_at
        })
        if self.is_first_slice:
            mime, _ = mimetypes.guess_type(self.name)
            data = [dict({
                'name': self.name,
                'mime': mime if mime is not None else '',
                'length': self.full_length
            })]
            json_metadata['data'] = data
        return json.dumps(json_metadata, separators=(',', ':'))


class Payload:
    meta: PayloadMetadata
    data: Bytes

    def __init__(self, meta: PayloadMetadata, data: Bytes) -> None:
        self.meta = meta
        self.data = data

    def __eq__(self, o) -> bool:
        return self.meta.slice_num == o.meta.slice_num

    def __le__(self, o) -> bool:
        return self.meta.slice_num < o.meta.slice_num

    def __gt__(self, o) -> bool:
        return self.meta.slice_num > o.meta.slice_num

    def __get_metadata_length(self, data: Bytes) -> int:
        for i, item in zip(range(len(data)), data):
            if item == 0:
                return i
        return -1

    @classmethod
    def extract_payload(cls, data: Bytes):
        metadata_length = cls.__get_metadata_length(cls, data)
        metadata = PayloadMetadata.from_json(
            data[:metadata_length].to_str(), metadata_length)
        data = data[metadata.start_at:metadata.end_at]
        return cls(metadata, data)

    def produce_payload(self, payload_id: str) -> Bytes:
        return Bytes.from_str(self.meta.to_json(payload_id))+[0]+self.data
